ADC set bandwidth_ 0
ADC/Param set oresv_rate_ 0

#  this example creates a three node topology
#
#                          n1
#                         /  \
#                    5Mb /    \ 5Mb 
#                       /      \    
#                      /  2Mb   \   
#                    n0----------n2 
#
#  and creates following sessions
#
#  1. a 2.5 Mbps session between n0 and n2 fid 1 (path taken n0-n1-n2)
#  2. a 1.5 Mbps session between n0 and n2 fid 2 (path taken n0-n2)
#  3. a 1.5 Mbps session between n0 and n2 fid 3 (path taken n0-n1-n2)
#  4. a 2.0 Mbps session between n0 and n2 fid 4 (unsuccessful as bandwith
#     not available so forwarded as best-effort fid 0)
 
# QOS extensions to LS routing add a instproc "get-explicit-route dest bw" 
# to Agent/rtProto/LS which returns explicit route to destination $dest
# with bandwith $bw available

# QOS Routing needs following two hooks from its environment
#
# 1. Link instproc get-bw{}
#    this should return the current available bandwith on the link 
#
# 2. Agent/rtProtoLS/bw-changed{}
#    the QOSR routing agent should be informed whenever the available 
#    bandwith on a link changes


# this example uses cbq links for fair sharing of bandwith and ADC/Param 
# for admission control

# the following procedure setups cbqlink between "n1" and "n2" with
# bandwith "bw" delay "delay". it also setups a default cbqclass for 
# normal traffic without bandwith requirements (with fid 0) with 
# "default" percent of bandwith available to it. it also creates an ADC/param
# instances for the link


Simulator instproc duplex-qosr-link { n1 n2 bw delay default} {

    # create cbq link
    $self duplex-link $n1 $n2 $bw $delay CBQ/WRR

    set link12 [$self link $n1 $n2]
    set link21 [$self link $n2 $n1]

    # root class
    $link12 set topclass [new CBQClass]
    [$link12 set topclass] setparams none 0 1 auto 8 2 0
    $link12 insert [$link12 set topclass]

    $link21 set topclass [new CBQClass]
    [$link21 set topclass] setparams none 0 1 auto 8 2 0
    $link21 insert [$link21 set topclass]

    # class for fid 0 traffic
    set defclass [new CBQClass]
    $defclass setparams [$link12 set topclass] true $default auto 1 1 0
    set queue1 [new Queue/DropTail]
    $queue1 set limit_ 20
    $defclass install-queue $queue1

    $link12 insert $defclass
    $link12 bind $defclass 0

    set defclass [new CBQClass]
    $defclass setparams [$link21 set topclass] true $default auto 1 1 0
    set queue1 [new Queue/DropTail]
    $queue1 set limit_ 20
    $defclass install-queue $queue1

    $link21 insert $defclass
    $link21 bind $defclass 0

    # admission control module
    $link12 set paramadc_ [new ADC/Param]
    $link21 set paramadc_ [new ADC/Param]
    [$link12 set paramadc_] set bandwidth_ $bw
    [$link21 set paramadc_] set bandwidth_ $bw
}

#implement get-bw function required for qos routing

# available bw = total bandwith * utilization - reserved bandwith
SimpleLink instproc get-bw {} {
    $self instvar paramadc_
    if {![info exists paramadc_]} {
	return 0
    }
    set bandwidth_ [bw_parse [$paramadc_ set bandwidth_]]
      return [expr $bandwidth_ * \
  		[$paramadc_ set utilization_] - \
  		[$paramadc_ set oresv_rate_] ]
}

# creates a session between $self and $dest of bandwith bw with fid $fid
# bind agaent $agnt to that session.
# 1. get-explicit-route
# 2. query admission control module along the path and make reservations
# 3. install cbq classes for that fid with that bandwith
# 
# in actual networks, steps 2 and 3 would have to be carried by a 
# signalling agent

Node instproc create-session {dest bw fid agnt} {
    $self instvar rtObject_
    $self instvar src_agent_

    if {![info exists rtObject_]} {
	return 0 ;# no rtObject_
    }

    $rtObject_ instvar rtProtos_

    if {![llength [array names rtProtos_ LS]]} {
	return 0 ;#no LS found
    }

    set rt $rtProtos_(LS)

    #get explicit-route

    set eroute [$rt get-explicit-route $dest $bw]
    if {![llength $eroute]} {
	puts "create session $bw $fid [$self id] - [$dest id] unsuccessful: bandwidth not available"
	$agnt set class_ 0
	return 0
    }

    #ok, now start making reservations along the path

    set sim [Simulator instance]

    set node1 $self

    set success 1 ;#whether reservation was successful
    set nlist "" ;#list of nodes where rollback has to be done in case

    foreach i $eroute {
	#get node from id
	set node2 [$sim get-node-by-id $i]
	#get link connecting node1 and node2
	set link [$sim link $node1 $node2]
	#get adc on them FIXME check whether it has adc
	set adc [$link set paramadc_]
	#reserve bandwith on link
	set ret [$adc admit_flow 0 $bw 0]
	#if ret false, rollback FIXME verify this
	if {!$ret} {
	    set success 0 ;#oops, we have to rollback
	    set rnode1 $self
	    foreach j $nlist {
		set rnode2 [$sim get-node-by-id $j]
		set rlink [$sim link $rnode1 $rnode2]
		set radc [$rlink set paramadc_]
		$radc rej_action 0 $bw 0
		set rnode1 $rnode2
	    }
	    break
	}
	lappend nlist $node2
	set node1 $node2
    } 
    
    if {!$success} {
	puts "create session $bw $fid [$self id] - [$dest id] unsuccessful: reservation unsuccessful"
	return 0 ;#sorry, failed to reserve
    }
    set node1 $self

    #so, reservation done; now set cbq parameters
    #and inform ls routing agent about bandwith changes

    $rt bw-changed

    foreach i $eroute {
	#get node from id
	set node2 [$sim get-node-by-id $i]

	#get ls routing agent on the node
	$node2 instvar rtObject_
	if {![info exists rtObject_]} {
	    return 0 ;# no rtObject_
	}
	$rtObject_ instvar rtProtos_
	if {![llength [array names rtProtos_ LS]]} {
	    return 0 ;#no LS found
	}
	set nrt $rtProtos_(LS)
	
	#call bw-changed
	$nrt bw-changed

	#get link connecting node1 and node2
	set link [$sim link $node1 $node2]
	set adc [$link set paramadc_]
	set cbqclass_ [new CBQClass]
	set cbqqueue_ [new Queue/DropTail]
	$cbqqueue_ set limit_ 20
	$cbqclass_ install-queue $cbqqueue_
	set ratio [expr [bw_parse $bw]/[bw_parse [$adc set bandwidth_]]]
	$cbqclass_ setparams [$link set topclass] true $ratio auto 1 1 0
	$link insert $cbqclass_
	$link bind $cbqclass_ $fid
	set node1 $node2
    } 

    #setup source routing

    $agnt set class_ $fid
    $agnt target $src_agent_

    eval "$src_agent_ install_connection $fid [$self id] [$dest id] [$self id] $eroute"
    
    puts "create session $bw $fid [$self id] - [$dest id] successful"
    return 1

}


#---------------------------------------------------------------------------
#---------------------------------------------------------------------------

set ns [new Simulator]

#enable source routing
$ns src_rting 1

$ns color 0 blue
$ns color 1 red
$ns color 2 green
$ns color 3 purple
$ns color 4 brown

set n0 [$ns node]
set n1 [$ns node]
set n2 [$ns node]

set f [open out.tr w]
$ns trace-all $f
set nf [open out.nam w]
$ns namtrace-all $nf

# create links
$ns duplex-qosr-link $n0 $n2 2Mb 2ms .1
$ns duplex-qosr-link $n0 $n1 5Mb 2ms .1
$ns duplex-qosr-link $n1 $n2 5Mb 2ms .1

#for nam simulation
$ns duplex-link-op $n0 $n1 orient right-up
$ns duplex-link-op $n0 $n2 orient right
$ns duplex-link-op $n1 $n2 orient right-down
$ns duplex-link-op $n0 $n1 queuePos 0.5

#create traffic sources

set udp0 [new Agent/UDP]
$ns attach-agent $n0 $udp0
set cbr0 [new Application/Traffic/CBR]
$cbr0 set rate_ 2.5Mb
$cbr0 attach-agent $udp0

set udp1 [new Agent/UDP]
$ns attach-agent $n0 $udp1
set cbr1 [new Application/Traffic/CBR]
$cbr1 set rate_ 1.5Mb
$cbr1 attach-agent $udp1

set udp2 [new Agent/UDP]
$ns attach-agent $n0 $udp2
set cbr2 [new Application/Traffic/CBR]
$cbr2 set rate_ 2.5Mb
$cbr2 attach-agent $udp2

set udp3 [new Agent/UDP]
$ns attach-agent $n0 $udp3
set cbr3 [new Application/Traffic/CBR]
$cbr3 set rate_ 3.0Mb
$cbr3 attach-agent $udp3

set null0 [new Agent/Null]
$ns attach-agent $n2 $null0

$ns connect $udp0 $null0
$ns connect $udp1 $null0
$ns connect $udp2 $null0
$ns connect $udp3 $null0

#enable link state routing
$ns rtproto LS

proc finish {} {
	global ns f nf
	$ns flush-trace
	close $f
	close $nf

	puts "running nam..."
	exec nam -r 20.6ms out.nam &
	exit 0
}


#
$ns at .6 "$n0 create-session $n2 2.5Mb 1 $udp0"
$ns at .7 "$cbr0 start"

$ns at .7 "$n0 create-session $n2 1.5Mb 2 $udp1"
$ns at .8 "$cbr1 start"

$ns at 1.1 "$n0 create-session $n2 1.5Mb 3 $udp2"
$ns at 1.2 "$cbr2 start"

$ns at 1.5 "$n0 create-session $n2 3.0Mb 4 $udp3"
$ns at 1.6 "$cbr3 start"

$ns at 4.0 "finish"

$ns run
